package kosoft.utils.redirectstdio;

// redirectstdio library to redirect java stdout/stderr to timestamped & size rolled file. 
// Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 

// This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
// Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
// the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
// Lesser General Public License for more details.


import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class RollingOutputStream extends OutputStream {


	public RollingOutputStream(String newPathPrefix) {
		filePathPrefix = newPathPrefix;
	}

	public RollingOutputStream(String newPathPrefix, long maxSize) {
		filePathPrefix = newPathPrefix;
		maxByteCount = maxSize;
	}


	File file = null;
	FileOutputStream fileOut = null;
	BufferedOutputStream subOut = null;
	int seqNum = 0;
	long byteCount = 0;

	String filePathPrefix;
	long maxByteCount = 4000000l;

	int numLogsToRetain = 10;
	public int getNumLogsToRetain() { return numLogsToRetain; }
	public void setNumLogsToRetain(int numLogsToRetain) { this.numLogsToRetain = numLogsToRetain; }

	boolean debug = false;
	public boolean getDebug() { return debug; }
	public void setDebug(boolean debug) { this.debug = debug; }
	

	/*
	 * These two need to match, since the format is what is used for a new file 
	 * and the pattern is what is checked for retaining old files.
	 */
	static SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd_HHmmss_SSS");
	static String dateFormatPattern = "\\d{8}_\\d{6}_\\d{3}";

	File getNextFile() {
		Date date = new Date();
		String filename = null;
		try { 
			if (debug) { 
				filename = filePathPrefix + "_" + "fixed" + ".log";
			}
			else { 
				filename = filePathPrefix + "_" + format.format(date) + ".log";
			}
		}
		catch (Exception e) {
		}
		
		// This was retained from earlier version where we were just numbering the files.

		if (filename == null) {
			int seq = 0;
			while (true) { 
				filename = filePathPrefix + "_" + Integer.toString(seq) + ".log";			   
				File file = new File(filename);
				if (!file.exists()) {
					break;
				}
				seq++;
			}

		}

		return new File(filename);
		
	}

	synchronized void createNewFile() {

		try {
			file = getNextFile();
			//System.out.println("rolling to " + filename);
			fileOut = new FileOutputStream(file);
			subOut = new BufferedOutputStream(fileOut);
			byteCount = 0;
		}
		catch (FileNotFoundException e) {
			throw new RuntimeException("File not found " + file.getAbsolutePath());
		}
		
		requestLogCheck();
	}

	
	synchronized void XcreateNewFile() {
		Date date = new Date();
		String filename = null;
		try { 
			filename = filePathPrefix + "_" + format.format(date) + ".log";
		}
		catch (Exception e) {
		}

		if (filename == null) {
			int seq = 0;
			while (true) { 
				filename = filePathPrefix + "_" + Integer.toString(seq) + ".log";			   
				File file = new File(filename);
				if (!file.exists()) {
					break;
				}
				seq++;
			}

		}

		try {
			//System.out.println("rolling to " + filename);
			fileOut = new FileOutputStream(filename);
			subOut = new BufferedOutputStream(fileOut);
			byteCount = 0;
		}
		catch (FileNotFoundException e) {
			throw new RuntimeException("File not found " + filename);
		}
	}

	void checkFileCreated() {
		if (subOut == null) {
			createNewFile();
		}
	}

	void checkRolloverFileOnFlush() throws IOException  {

		if (byteCount > maxByteCount) {

			// If not null then we are rolling the file.
			if (subOut != null) {

				subOut.flush();
				subOut.close();
				fileOut.flush();
				fileOut.close();
				subOut = null;
				fileOut = null;
			}
		}

		if (subOut == null) {
			createNewFile();
		}

	}

	Thread logChecker = null;

	boolean requestLogCheck = false;

	synchronized void requestLogCheck() {
		
		// If is zero then we keep all logs, so we don't need to bother starting up the bg thread.
		if (getNumLogsToRetain() <= 0 ) {  
			return;
		}

		// Request a check.
		requestLogCheck = true;
		
		// Startup background thread, if it not already started.
		if (logChecker == null) {
			
			logChecker = new Thread() {
				public void run() {
					runLogCheckerBg();
				}
			};
			
			logChecker.setDaemon(true);
			logChecker.start();

		}
	}

	synchronized boolean checkRequstLogCheck() {
		
		// If someone has requested a check, clear it and return true.
		if (requestLogCheck) {
			requestLogCheck = false;
			return true;
		}
		
		// Since no check required, we can stop the bg thread - next request will start it again.
		else {
			logChecker = null;
			return false;
		}
	}

	void runLogCheckerBg() {
		do { 
			checkRetainedLogs();
		}
		while (checkRequstLogCheck());
	}
	
	void checkRetainedLogs() {
		
		try {
			
			File prefixFile = new File (filePathPrefix);

			File parentDir = null;
			String filePrefix = "";

			// In workign out parent, we use the "filePathPrefix" and bust it up into 
			// the "parent" dir and the remaining as the filePrefix.  We also cater for case
			// (not recommended) where only parent dir is specificed in the pathPrefix. 
			if (prefixFile.isDirectory() && filePathPrefix.endsWith("/")) {
				parentDir = prefixFile;
			}
			else {
				parentDir = prefixFile.getParentFile();
				filePrefix = prefixFile.getName();
			}

			// When no parent dir is specified (mainly this is in testing) 
			if (parentDir == null) {
				parentDir = new File(".");
			}

			// Get all files in current Directory that match the pattern.

			String formString = Pattern.quote(filePrefix) + "_" + dateFormatPattern + Pattern.quote(".log");
			final Pattern pattern = Pattern.compile(formString);

			File[] files = parentDir.listFiles(new FilenameFilter() {

				@Override
				public boolean accept(File dir, String name) {
					Matcher match = pattern.matcher(name);

					boolean found = match.find();
					return found;
				}

			});


			int numToKeep = getNumLogsToRetain();

			int numToRemove = files.length - numToKeep;

			// Check if we actually have any work to do.
			if (numToRemove <= 0) {
				return; 
			}

			// Sort then so they are in alphabetical order. 
			Arrays.sort(files); 

			// Then delete the oldest ones. 
			for (int i=0; i < numToRemove; i++) {
				File file = files[i];
				file.delete();
			}
		}
		catch (Throwable e) {
			System.err.println("Problem deleting old log files " + e.getMessage());
			e.printStackTrace(System.err);
		}
	}

	public void close()  throws IOException  {
		if (subOut != null) {
			subOut.flush();
			subOut.close();
			fileOut.flush();
			fileOut.close();
			subOut = null;
			fileOut = null;
		}
	}

	public void flush ()  throws IOException  {

		//System.out.println("flush");
		checkRolloverFileOnFlush();
		if (subOut != null) {
			subOut.flush();
		}
	}

	public void write(byte[] b)  throws IOException  {
		checkFileCreated();
		byteCount += b.length;
		subOut.write(b);
	}

	public void write(byte[] b, int off, int len)  throws IOException {
		checkFileCreated();
		byteCount += len;
		subOut.write(b, off, len);
	}

	public void write(int b)  throws IOException  {
		checkFileCreated();
		byteCount += 1;
		subOut.write(b);
	}

}